"""Reading from __init__.py"""

from . import constants
from . import cores
from . import profiles
from . import results
from . import shapes
from . import utils

__all__ = ["constants", "cores", "profiles", "results", "shapes", "utils"]


from pathlib import Path
from typing import NamedTuple
import astropy.units as u
import numpy as np


class SpxGeometry(NamedTuple):
    nav: int
    lat: float
    lon: float
    phase: float
    emission: float
    azimuth: float
    wgeom: float
    wavelengths: np.ndarray
    spectrum: np.ndarray
    error: np.ndarray


class SpxFileData(NamedTuple):
    fwhm: float
    lat: float
    lon: float
    geometries: tuple[SpxGeometry, ...]


def read(path: str) -> SpxFileData:
    """
    Read in a .spx file from disk.

    The returned data is a SpxFileData object containing the data from the .spx file.
    For example, you can access the latitude of the observation with `data.lat` or the
    spectrum of the first geometry with `data.geometries[0].spectrum`.

    Args:
        path: The path to the .spx file.

    Returns:
        A SpxFileData object of the form `(fwhm, lat, lon, geometries)` containing the
        data from the .spx file.
    """
    with open(path, 'r') as f:
        fwhm, hdr_lat, hdr_lon, ngeom = (float(x) for x in f.readline().split())
        ngeom = int(ngeom)
        geometries: list[SpxGeometry] = []
        for _ in range(ngeom):
            nconv = int(float(f.readline()))
            nav = int(float(f.readline()))
            lat, lon, phase, emission, azimuth, wgeom = (
                float(x) for x in f.readline().split()
            )
            wavelengths = np.full(nconv, np.nan)
            spectrum = np.full(nconv, np.nan)
            error = np.full(nconv, np.nan)
            for i in range(nconv):
                w, s, e = (float(x) for x in f.readline().split())
                wavelengths[i] = w
                spectrum[i] = s
                error[i] = e
            geometries.append(
                SpxGeometry(
                    nav=nav,
                    lat=lat,
                    lon=lon,
                    phase=phase,
                    emission=emission,
                    azimuth=azimuth,
                    wgeom=wgeom,
                    wavelengths=wavelengths,
                    spectrum=spectrum,
                    error=error,
                )
            )
    return SpxFileData(
        fwhm=fwhm,
        lat=hdr_lat,
        lon=hdr_lon,
        geometries=tuple(geometries),
    )


def convert_MJysr_to_Wcm2srum(
    wavelengths: np.ndarray,
    spectrum: np.ndarray,
) -> np.ndarray:
    """
    Convert a spectrum or cube from MJy/sr to W/cm2/sr/micron.
    """
    while len(wavelengths.shape) < len(spectrum.shape):
        # work for cubes passed to sp
        wavelengths = np.expand_dims(wavelengths, -1)

    spx_MJysr = spectrum * u.MJy / u.sr
    spx_Wcm2srum = spx_MJysr.to(
        u.W / (u.cm * u.cm) / u.sr / u.micron,
        equivalencies=u.spectral_density(wavelengths * u.micron),
    )

    return spx_Wcm2srum.value


def convert_Wcm2srum_to_MJysr(
    wavelengths: np.ndarray,
    spectrum: np.ndarray,
) -> np.ndarray:
    """
    Convert a spectrum or cube from W/cm2/sr/micron to MJy/sr.
    """
    while len(wavelengths.shape) < len(spectrum.shape):
        # work for cubes passed to sp
        wavelengths = np.expand_dims(wavelengths, -1)

    spx_Wcm2srum = spectrum * u.W / (u.cm * u.cm) / u.sr / u.micron
    spx_MJysr = spx_Wcm2srum.to(
        u.MJy / u.sr,
        equivalencies=u.spectral_density(wavelengths * u.micron),
    )
    return spx_MJysr.value


    """Write data (in MJy/sr) to a .spx file (in W/cm2/sr/um)."""
    converted_spectrum = convert_MJysr_to_Wcm2srum(wavelengths, spectrum)
    converted_error = convert_MJysr_to_Wcm2srum(wavelengths, error)
    spx_data = construct_spx(wavelengths, converted_spectrum, converted_error, **kwargs)
    write(filepath, spx_data)


"""This module contains the classes for creating Profile objects."""

from itertools import zip_longest
from collections import defaultdict

from . import constants
from . import shapes
from . import utils


class Profile:
    def __init__(self, label=None):
        """This is the base class for all Profile objects. It should never be instantiated directly, use a subclass such as ``GasProfile`` or ``TemperatureProfile``"""
        self.retrieved = False
        self.core = None
        self.label = label

    def _add_result(self, df):
        """Take in a DataFrame created by reading in the .mre file (results.NemesisResult.read_mre) and assign the correct attributes
        
        Args:
            df: pandas.DataFrame with columns "prior", "prior_error", "retrieved", "retrieved_error" and a row for each parameter
            
        Returns:
            None"""
        
        # Get order of parameters
        names = self.shape.NAMES
        assert len(names) == len(df)
        # Set attributes of the child Shape object
        for name, (_, row) in zip(names, df.iterrows()):
            for title, value in zip(row.index, row.values):
                if "prior" in title:
                    attrname = title.replace("prior", name)
                elif "retrieved" in title:
                    attrname = title.replace("retrieved", f"retrieved_{name}")
                setattr(self.shape, attrname, value)

        # Toggle retrieved flag
        self.retrieved = True

    def __str__(self):
        # generate table title
        try:
            self.label
        except:
            self.label = None
        if self.label is not None:
            title = self.label
        elif isinstance(self, GasProfile):
            title = self.gas_name
        elif isinstance(self, (AerosolProfile, ImagRefractiveIndexProfile)):
            title = self.__class__.__name__ + f" ID={self.aerosol_id}"
        else:
            title = self.__class__.__name__

        # get shape attributes
        attrs = self.shape.__dict__.keys()

        # group attributes together by parameter
        grouped = defaultdict(list)
        for attr in attrs:
            base = attr.removeprefix("retrieved_").removesuffix("_error")
            grouped[base].append(attr)

        # extract all the attribute values
        values = []
        for name, attrs in grouped.items():
            values.append([])
            for i, attr in zip_longest(range(4), attrs):
                if attr is not None:
                    values[-1].append(getattr(self.shape, attr))
                else:
                    values[-1].append("NA")
    
        return utils.generate_ascii_table(title, ["Prior", "Error", "Retrieved", "Error"], grouped.keys(), values)


class TemperatureProfile(Profile):
    def __init__(self, filepath, **kwargs):
        """Create a temperature profile from a prior file
        
        Args:
            filepath: The filepath of the prior temperature profile
            label: (optional) An arbitrary label to associate with this profile
        """
        super().__init__(**kwargs)
        self.shape = shapes.Shape0(filepath=filepath)

    def __repr__(self):
        return f"<TemperatureProfile [{self.create_nemesis_string()}]>"

    def create_nemesis_string(self):
        """Create the NEMESIS code that represents the temperature profile. Temperature profiles currently only support mode 0 0 0 so this function's return value is always constant
        
        Args:
            None
            
        Returns:
            str: '0 0 0'"""
        return f"0 0 0"
    
    def generate_apr_data(self):
        """Generate the section of the .apr file for this temperature profile
        
        Args:
            None
            
        Returns:
            str: The string to write to the .apr file"""
        return "0 0 0 - Temp\n" + self.shape.generate_apr_data()

    def _add_result(self, df):
        self.shape.data = df
        self.retrieved = True


class GasProfile(Profile):
    def __init__(self, gas_name=None, gas_id=None, isotope_id=0, shape=None, **kwargs):
        """Create a profile for a given gas (optionally an isotopologue) with a given shape
        
        Args:
            gas_name: The name of the gas (eg 'CH4'). Specify either this OR gas_id
            gas_id: The radtrans ID of the gas (eg. 6). Specify either this OR gas_name
            isotope_id: The ID of the isotopologue to use. Use 0 for a mix of all at terrestrial abundance
            shape: A Shape object to use for the profile shape
            label: (optional) An arbitrary label to associate with this profile"""
        super().__init__(**kwargs)
        self.isotope_id = isotope_id
        if shape is None:
            raise ValueError("shape attribute must be specified")
        self.shape = shape
        if not ((gas_id is None) ^ (gas_name is None)):
            raise ValueError("Specifiy exactly one of gas_name or gas_id (not both!)")
        if gas_name is None:
            self.gas_id = gas_id
            self.gas_name = constants.GASES.loc[constants.GASES.radtrans_id == gas_id].name.iloc[0]
        elif gas_id is None:
            self.gas_name = gas_name
            self.gas_id = constants.GASES.loc[constants.GASES.name == gas_name].radtrans_id.iloc[0]

    def __repr__(self):
        return f"<GasProfile {self.gas_name} [{self.create_nemesis_code()}]>"

    def create_nemesis_code(self):
        """Create the NEMESIS code that represents the gas profile (eg. 23 0 1)
        
        Args:
            None
            
        Returns:
            str: The NEMESIS code"""
        return f"{self.gas_id} {self.isotope_id} {self.shape.ID}"
    
    def generate_apr_data(self):
        """Generate the section of the .apr file for this gas profile
        
        Args:
            None
            
        Returns:
            str: The string to write to the .apr file"""
        return self.create_nemesis_code() + " - " + self.gas_name + "\n" + self.shape.generate_apr_data()


class AerosolProfile(Profile):
    def __init__(self, shape, **kwargs):
        """Create a profile for a given aerosol with a given shape
        
        Args:
            aerosol_id: The ID of the aerosol
            shape: A Shape object to use for the profile shape
            label: (optional) An arbitrary label to associate with this profile"""
        super().__init__(**kwargs)
        self.shape = shape

    def __repr__(self):
        return f"<AerosolProfile {self.aerosol_id} [{self.create_nemesis_code()}]>"

    def create_nemesis_code(self):
        """Create the NEMESIS code that represents the aerosol profile (eg. -1 0 32)
        
        Args:
            None
            
        Returns:
            str: The NEMESIS code"""
        return f"-{self.aerosol_id} 0 {self.shape.ID}"
    
    def generate_apr_data(self):
        """Generate the section of the .apr file for this aerosol profile
        
        Args:
            None
            
        Returns:
            str: The string to write to the .apr file"""
        return self.create_nemesis_code() + f" - CP{self.aerosol_id}\n" + self.shape.generate_apr_data()


class ImagRefractiveIndexProfile(Profile):
    def __init__(self, shape, **kwargs):
        """Create a profile that retrieves the imaginary part of a clouds refractive index. Will require a corresponding AerosolProfile
        to be defined. This is currently limited to constant refractive index as a function of wavelength.

        Args:
            aerosol_id (int): The ID of the aerosol
            shape (Shape444): The associated Shape444 object containing the aerosol properties
            label: (optional) An arbitrary label to associate with this profile
            """

        super().__init__()
        assert isinstance(shape, shapes.Shape444)
        self.shape = shape

    def __repr__(self):
        return f"<ImagRefractiveIndexProfile [{self.create_nemesis_string()}]>"

    def create_nemesis_string(self):
        """Create the NEMESIS code that represents the profile.
        
        Args:
            None
            
        Returns:
            str: The NEMESIS string"""
        return f"444 {self.aerosol_id} 444"
    
    def generate_apr_data(self):
        return f"{self.create_nemesis_string()} - n{self.aerosol_id}\n{self.shape.generate_apr_data()}"



def read_between_lines(file, start, end):
    """Take in an open file object and get the data between line numbers 'start' and 'end'.
    It is inclusive of start and exclusive of end (ie. start <= line < end)
    
    Args:
        file: File-like object to read from
        start: Line number to start reading from
        end: Line number to stop reading at
        
    Returns:
        str: File data
    """
    file.seek(0)
    lines = []
    for i, line in enumerate(file):
        if start <= i < end:
            lines.append(line)
    return "".join(lines)


def indent(string, level=1, spaces_per_level=4):
    """Take in a string and return the same string indented at each line break.
    
    Args:
        string: The string to indent
        level: The level at which to indent
        spaces_per_level: How many spaces to add per level of indentation
        
    Returns:
        str: The indented string
    """
    lines = string.split("\n")
    ind = " "*spaces_per_level*level
    return "\n".join([ind+l for l in lines])


def write_nums(file, *nums, sep="    ", dp=6):
    for i, n in enumerate(nums):
        file.write(f"{n:.{dp}f}")
        if i != len(nums)-1:
            file.write(sep)
        else:
            file.write("\n")


def generate_ascii_table(title, headers, row_headers, rows):
    """Create a text-based table with a header column using Unicode box drawing characters.

    Args:
        headers (list): A list of column headers.
        row_headers (list): A list of row header labels (one for each row).
        rows (list of list): A list of rows, where each row is a list of column values.
        top_left (str): The string to place in the top-left cell (default is empty).

    Returns:
        str: The formatted table as a string.
    """
    # Add the top-left string to the headers
    headers = [title] + headers

    # Determine the column widths, including the row header column
    num_columns = len(headers)
    col_widths = [max(len(str(row[i])) for row in rows) for i in range(len(headers) - 1)]
    col_widths.insert(0, max(len(str(h)) for h in row_headers))  # Row header column
    col_widths = [max(col_widths[i], len(headers[i])) for i in range(num_columns)]

    # Box drawing characters
    top_left_corner = "┌"
    top_right_corner = "┐"
    bottom_left_corner = "└"
    bottom_right_corner = "┘"
    horizontal = "─"
    vertical = "│"
    cross_top = "┬"
    cross_middle = "┼"
    cross_bottom = "┴"

    # Helper to create rows
    def make_row(cells, left, middle, right):
        return f"{left} " + f" {middle} ".join(f"{str(cell):<{col_widths[i]}}" for i, cell in enumerate(cells)) + f" {right}"

    # Header separator
    def make_separator(left, middle, right):
        return f"{left}" + f"{middle}".join(horizontal * (col_widths[i] + 2) for i in range(num_columns)) + f"{right}"

    # Build the table
    table = [
        make_separator(top_left_corner, cross_top, top_right_corner),  # Top border
        make_row(headers, vertical, vertical, vertical),  # Header row
        make_separator("├", cross_middle, "┤"),  # Separator under header
    ]
    for row_header, row in zip(row_headers, rows):
        table.append(make_row([row_header] + row, vertical, vertical, vertical))
    table.append(make_separator(bottom_left_corner, cross_bottom, bottom_right_corner))  # Bottom border

    return "\n".join(table)


"""This module contains any constants used across the module. These attributes are not intended to be used from outside the module but may be useful in some circumstances."""

import pandas as pd
import os


def _read_radtrans_gas_id():
    data = []
    with open(PATH+"data/radtrans_ids.csv") as file:
        for line in file:
            x = line.rstrip("\n").split(",")
            data.append(x[:7] + [x[7:],])
    df = pd.DataFrame(data)
    df.columns = ["radtrans_id", "name", "H04", "G03", "H12", "H16", "N_iso", "isotopes"]
    df["G03"].astype(int)
    df = df.astype({"radtrans_id":int, "name": str, "H04": int, "G03": int, "H12": int, "H16": int, "N_iso": int})
    return df

PATH = os.path.dirname(__file__) + "/" #: The absolute path of the prepackaged data directory

GASES = _read_radtrans_gas_id() #: The database of gas names and IDs used by NEMESIS

DISTANCES = {"jupiter": 5.2,
             "saturn": 9.546,
             "uranus": 19.2,
             "neptune": 30.0,
             "titan": 9.546,} #: The distnaces from the Suyn for all the bodies with significant atmospheres


""" 
Each Shape object represents a differnet model in NEMESIS. Eg. ``Shape1`` represents profile shape 1 which is a constant VMR 
up to a knee pressure then a fractional scale height. For a ful list of possible models see the NEMESIS manual. Please note: 
due to the large number of profile shapes NEMESIS supports, 

Adding a Shape subclass
=======================

All subclasses of Shape must implement:
    - ID: Integer ID of the shape in NEMESIS
    - NAMES: The names of the parameters in order they appear in the mre file (do not include error parameters here)
    - __init__: Instantiates the parent class and defines:
        - nemesis_code: The number of the profile in the NEMESIS manual
        - Any parameters associated with the model (eg. knee pressure, fsh etc). These MUST be defined in the order that they appear in the .apr file
    - generate_apr_data: Returns a string of the profile shape parameters in the format
                         that the .apr file expects.

Additionally, some shapes require additional files in the core (eg. tempapr.dat). If this is the case,
override the create_required_files method.

Finally, add the new ShapeN class to the ALL_SHAPES list at the end of the file.

Class template::

    @shapeclass
    class ShapeN(Shape):
        "copy the desciption from the NEMESIS manual here (use triple quotes!)"
        ID: ClassVar[int] = N
        NAMES:  ClassVar[list[str]] = ["arg_name_1", "arg_name_2", ...]
        arg_name_1: type_1
        arg_name_2: type_2
        arg_name_2_error: type_2
        ...

        def generate_apr_data(self):
            return a string for apr file

        # optionally
        def create_required_files(self, directory):
            shutil.copy(..., directory)

"""

from dataclasses import dataclass
from typing import ClassVar
import shutil

from . import spx, utils


def shapeclass(*args, **kwargs):
    """Syntatic sugar for ``@dataclass(repr=False)``"""
    return dataclass(*args, **kwargs, repr=False)
    


class Shape:
    def __init__(self):
        """Do not instantiate directly, use a subclass"""
        pass

    def __repr__(self):
        out = f"Shape:\n    ID: {self.ID}"
        for name, value in self.__dict__.items():
            out += f"\n    {name}: {value}"    
        return out
    
    def generate_apr_data(self):
        """Generate the section of the .apr file that stores the shape parameters
        
        Args:
            None
            
        Returns:
            str: The string to write to the .apr file"""
        pass

    def create_required_files(self, directory):
        """Some Shapes require additional files to be created/copied into the core directory"""
        pass

@shapeclass
class Shape0(Shape):
    """Profile is to be treated as continuous over the pressure range of runname.ref, the
    next line of the .apr file should then contain a filename, which specifies the a
    priori profile as a function of height and should have the same number of levels
    as the .ref file."""
    ID: ClassVar[int] = 0
    NAMES: ClassVar[list[str]] = ["None"]
    filepath: str

    def create_required_files(self, directory):
        shutil.copy(self.filepath, directory)

    def generate_apr_data(self):
        return self.filepath.split("/")[-1]

@shapeclass
class Shape1(Shape):
    """Profile is to be represented as a deep VMR up to a certain ‘knee’ pressure, and
then a defined fractional scale height. The next line of the .apr file then contains
the ‘knee’ pressure, followed by the a priori deep and fractional scale height
values together with their estimated errors. """
    ID: ClassVar[int] = 1
    NAMES: ClassVar[list[str]] = ["deep_vmr", "fsh"]
    knee_pressure: float
    deep_vmr: float
    deep_vmr_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.knee_pressure}\n{self.deep_vmr} {self.deep_vmr_error}\n{self.fsh} {self.fsh_error}"

@shapeclass
class Shape2(Shape):
    """Profile is to be represented by a simple scaling of the corresponding profile
runname.ref (for T, v.m.r.), aerosol.ref (for aerosol density), parah2.ref (for
para-H2 fraction) or fcloud.ref (for fractional cloud cover). The next line of the
.apr file then contains the a priori factor and error."""
    ID: ClassVar[int] = 2
    NAMES: ClassVar[list[str]] = ["scale_factor"]
    scale_factor: float
    scale_factor_error: float

    def generate_apr_data(self):
        return f"{self.scale_factor} {self.scale_factor_error}"

@shapeclass
class Shape4(Shape):
    """Very similar to Shape1 in that the profile is to be
represented as a deep value up to a certain 'knee' pressure, and then a defined
fractional scale height. However, in this case the knee pressure is also a variable
parameter and thus must be supplied with an error estimate."""
    ID: ClassVar[int] = 4
    NAMES: ClassVar[list[str]] = ["knee_pressure", "deep_vmr", "fsh"]
    knee_pressure: float
    knee_pressure_error: float
    deep_vmr: float
    deep_vmr_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.knee_pressure} {self.knee_pressure_error}\n{self.deep_vmr} {self.deep_vmr_error}\n{self.fsh} {self.fsh_error}"

@shapeclass
class Shape32(Shape):
    """Similar to model 8 in that profile is a cloud profile represented by a variable
base pressure, specific density at the level and fractional scale height. The next
line of the .apr file then contains the a priori base pressure, followed by the a
priori opacity and fractional scale height values together with their estimated
errors. All quantities are taken as logs so negative fractional scale heights are
not allowed. Difference from Model 8 is that cloud density at pressures greater
than the base pressure is set to drop exponentially with increasing pressure with
a scale height of 1km, rather than just being set to zero. This makes it easier for
NEMESIS to actually find an optimal value of the knee pressure."""
    ID:  ClassVar[int] = 32
    NAMES: ClassVar[list[str]] = ["opacity", "fsh", "base_pressure"]
    base_pressure: float
    base_pressure_error: float
    opacity: float
    opacity_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.base_pressure} {self.base_pressure_error}\n{self.opacity} {self.opacity_error}\n{self.fsh} {self.fsh_error}"
        
@shapeclass
class Shape48(Shape):
    """As model 32, in that profile is a cloud profile represented by a variable base
pressure, specific density at the level, fractional scale height, but also a variable
top pressure. The next line of the .apr file then contains the a priori base
pressure, followed by the a priori top pressure, opacity and fractional scale
height values together with their estimated errors."""
    ID:  ClassVar[int] = 48
    NAMES: ClassVar[list[str]] = ["base_pressure", "top_pressure", "fsh", "opacity"]
    base_pressure: float
    base_pressure_error: float
    top_pressure: float
    top_pressure_error: float
    opacity: float
    opacity_error: float
    fsh: float
    fsh_error: float

    def generate_apr_data(self):
        return f"{self.base_pressure} {self.base_pressure_error}\n{self.top_pressure} {self.top_pressure_error}\n{self.opacity} {self.opacity_error}\n{self.fsh} {self.fsh_error}"
        
@shapeclass
class Shape444(Shape):
    """If VARIDENT(IVAR,1) is equal to 444 then the parameter described is a retrieval
of the imaginary part of a cloud’s complex refractive index spectrum. The cloud
particle identifier is given by VARIDENT(IVAR,2). The next line contains the
name of a separate input file, which contains the following information. Line 1
contains the mean radius of the particle size distribution and error (assumes standard
size distribution), while line 2 gives the variance of the size distribution and error.
Line 3 gives the number of wavelengths/wavenumbers for which imaginary
refractive index spectrum is tabulated, together with the correlation length of this a
priori spectrum. Line 4 gives a reference wavelength/wavenumber and the value of
the real part of the refractive index at that reference. Line 5 gives the
wavelength/wavenumber to which the extinction cross-section spectrum should be
normalised. Following lines contain the wavelengths/wavenumbers and the a priori
values of the imaginary refractive index spectrum and errors. In this model, the code
the real part of the refractive index spectrum is calculated with a Kramers-Kronig
analysis and then the Mie scattering properties of the particles calculated."""
    ID: ClassVar[int] = 444
    NAMES: ClassVar[list[str]] = ["radius", "variance", "refractive_index"]
    radius: float
    radius_error: float
    variance: float
    variance_error: float
    refractive_index: complex
    refractive_index_error: float

    def generate_apr_data(self):
        return f"cloudf{self.aerosol_id}.dat"

    def create_required_files(self, directory):
         # Get first wavelength
        wl = spx.read(directory+"nemesis.spx").geometries[0].wavelengths[0]

        # Get number of wavelengths in xsc file
        with open(directory+"nemesis.xsc") as file:
            lines = file.read().split("\n")
            nwave = (len(lines)-2) // 2
            refwave = float(lines[1].split()[0])
        
        # Generate cloudfN.dat
        with open(directory + f"cloudf{self.aerosol_id}.dat", mode="w+") as file:
            utils.write_nums(file, self.radius, self.radius_error)
            utils.write_nums(file, self.variance, self.variance_error)
            file.write(f"{nwave}    -1\n")
            utils.write_nums(file, refwave, self.refractive_index.real)
            utils.write_nums(file, refwave)
            for line in lines:
                vals = line.split()
                if len(vals) < 2:
                    continue
                print(float(vals[0]), repr(self.refractive_index.imag), repr(self.refractive_index_error))
                utils.write_nums(file, float(vals[0]), self.refractive_index.imag, self.refractive_index_error)
    

def get_shape_from_id(id_):
    """Given a shape ID integer, return a reference to the class corresponding to that ID. Note that this returns a class, not an instantiated object."""
    id_ = int(id_)
    for shape_class in ALL_SHAPES:
        if shape_class.ID == id_:
            return shape_class


ALL_SHAPES = [Shape0, Shape1, Shape2, Shape4, Shape32, Shape48, Shape444] #: A list of all the Shape classes


import os
import shutil
import pandas as pd
import pickle
import time
import warnings

from . import constants
from . import spx
from . import profiles as profiles_ # to avoid namespace shadowing by NemesisCore.profiles


class NemesisCore:
    core_id = 0

    def __init__(self, 
                 parent_directory, 
                 spx_file, 
                 ref_file=None, 
                 profiles=list(), 
                 planet="jupiter", 
                 scattering=True, 
                 forward=False,
                 num_iterations=30,
                 num_layers=120, 
                 bottom_layer_height=-80, 
                 instrument="NIRSPEC", 
                 fmerror_factor=1,
                 fmerror_pct=None,
                 fmerror_value=None,
                 cloud_cover=1.0):
        """Create a NEMESIS core directory with a given set of profiles to retrieve
        
        Args:
            parent_directory: The directory in which to create the core folder 
            spx_file: Path to the spectrum file to fit
            ref_file: Path to the ref file to use (if left blank it will use the default for that planet)
            profiles: List of profiles.Profile objects to retrieve
            planet: Name of the planet being observed. Must be one of 'jupiter', 'saturn', 'uranus', 'neptune' or 'titan'.
            scattering: Whether to run a scattering retrieval or not
            forward: Whether of not to run a forward model (ie. set number of iterations = 0)
            num_iterations: Number of iterations to run in the retrieval (if forward is set this has no effect)
            num_layers: The number of atmospheric layers to simulate
            bottom_layer_height: The height in km of the bottom layer of the atmosphere
            instrument: Either 'NIRSPEC' or 'MIRI'; determines which set of ktables to use
            fmerror_factor: The factor by which to multiply the error on the spectrum (see also, fmerror_pct and fmerror_value)
            fmerror_pct: If given, instead of using fmerror_factor or fmerror_value, use a flat percentage of the brightness (eg. 0.1 = 10%) (see also, fmerror_factor and fmerror_value)
            fmerror_value: If given, instead of using fmerror_factor or fmerror_pct, use a flat value in W/cm2/sr/um (see also, fmerror_factor and fmerror_pct)
            cloud_cover: If scattering, then this is the fractional cloud cover between 0 and 1 (should not usually be changed)
        """
        # Increment the global core counter
        NemesisCore.core_id += 1

        # Assign attributes passed in
        self.spx_file = spx_file
        self.profiles = profiles
        self.planet = planet.lower()
        self.scattering = scattering
        self.forward = forward
        self.num_iterations = num_iterations
        self.num_layers = num_layers
        self.bottom_layer_height = bottom_layer_height
        self.instrument = instrument
        self.fmerror_factor = fmerror_factor
        self.fmerror_pct = fmerror_pct
        self.fmerror_value = fmerror_value
        self.cloud_cover = cloud_cover

        # Set the directories of the parent folder and own cores
        self.parent_directory = parent_directory
        self.directory = parent_directory + f"core_{self.core_id}/"

        # Check if we are perfoming a scattering run with more than 39 layers (NEMESIS hates this...)
        if num_layers > 39 and self.scattering:
            warnings.warn(f"Too many atmospheric layers specified for a scattering run ({num_layers} vs. 39). Automatically reducing to 39")
            self.num_layers = 39

        # Create the directory tree if it doesn't already exist and clear it if it does
        os.makedirs(self.parent_directory, exist_ok=True)
        if os.path.exists(self.directory):
            shutil.rmtree(self.directory)
        os.makedirs(self.directory)

        # Set ref file if not specified:
        if ref_file is None:
            self.ref_file = constants.PATH + f"data/{planet}/{planet}.ref"
            warnings.warn(f"No ref file specified. Using the default in {self.ref_file}")
        else:
            self.ref_file = ref_file
        
        # Parse the ref file:
        self.ref = parse_ref_file(self.ref_file)

        # Set number of aerosol modes (incremented by add_aerosol_mode)
        self.num_aerosol_modes = 0

        # Add a reference to self in each Profile and check that there are no Aerosol profiles
        for profile in self.profiles:
            if isinstance(profile, (profiles_.AerosolProfile, profiles_.ImagRefractiveIndexProfile)):
                raise ValueError("Aerosol-related profile specified in constructor - please use core.add_aerosol_profile(...) instead!")
            profile.core = self

        # If in forward mode, set the number of iterations to 0
        if self.forward:
            self.num_iterations = 0

        # Raise an error if trying to use features not implemented yet
        if planet != "jupiter":
            raise Exception("Eleos does not support planets other than Jupiter yet! ")
        
        # Copy in the boilerplate files
        self._copy_input_files()
        self._copy_template_files()

    def __str__(self):
        return f"<NemesisCore: {self.directory}>"

    def _save_core(self):
        with open(self.directory+"core.pkl", mode="wb") as file:
            pickle.dump(self, file)

    def _add_profile(self, profile):
        profile.core = self
        self.profiles.append(profile)

    def _copy_input_files(self):
        """Copy the given .spx and .ref file into the core
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            nemesis.spx
            nemesis.ref"""

        shutil.copy(self.ref_file, self.directory+"nemesis.ref")
        shutil.copy(self.spx_file, self.directory+"nemesis.spx")

    def _copy_template_files(self):
        shutil.copy(constants.PATH + "data/statics/nemesis.cia", self.directory)
        shutil.copy(constants.PATH + "data/statics/nemesis.abo", self.directory)
        shutil.copy(constants.PATH + "data/statics/nemesis.nam", self.directory)
        shutil.copy(constants.PATH + "data/statics/nemesis.sol", self.directory)
        shutil.copy(constants.PATH + "data/statics/makephase.inp", self.directory)
        shutil.copy(constants.PATH + f"data/{self.planet}/parah2.ref" , self.directory)

    def _generate_inp(self):
        """Generate the nemesis input file

        Args:
            num_iterations: Number of iterations to run (has no effect if NemesisCore.forward is True)
        
        Returns:
            None
            
        Creates:
            nemesis.inp"""
        with open(constants.PATH+"data/statics/nemesis.inp") as file:
            out = file.read()
            out = out.replace("<SCATTERING>", str(int(self.scattering)))
            out = out.replace("<N_ITERATIONS>", str(self.num_iterations))
        with open(self.directory+"nemesis.inp", mode="w+") as file:
            file.write(out)

    def _generate_apr(self):
        """Generate the nemesis.apr file from the profile list

        Args:
            None
        
        Returns:
            None
            
        Creates:
            nemesis.apr"""
        out = f"*******Apriori File*******\n           {len(self.profiles)}\n"
        for profile in self.profiles:
            profile.shape.create_required_files(self.directory)
            out += profile.generate_apr_data() + "\n"
        with open(self.directory + "nemesis.apr", mode="w") as file:
            file.write(out)

    def _generate_set(self):
        """Generate the settings file for NEMESIS
        
        Args:
            None

        Returns:
            None
            
        Creates:
            nemesis.set"""

        with open(constants.PATH+"data/statics/template.set", mode="r") as file:
            out = file.read()
        out = out.replace("<DISTANCE>", f"{constants.DISTANCES[self.planet]:.3f}")
        out = out.replace("<SUNLIGHT>", f"{int(self.scattering)}")
        out = out.replace("<BOUNDARY>", f"{int(self.scattering)}")
        out = out.replace("<N_LAYERS>", f"{int(self.num_layers)}")
        out = out.replace("<BASE>", f"{self.bottom_layer_height:.2f}")
        with open(self.directory+"nemesis.set", mode="w+") as file:
            file.write(out)

    def _generate_flags(self, inormal=0, iray=1, ih2o=0, ich4=0, io3=0, inh3=0, iptf=0, imie=0, iuv=0):
        """Generate the flags file. As a general rule, leave all this at the defaults. The descriptions are copied directly fron the NEMESIS manual.
        Args:
            inormal: 0 or 1 depending on whether the ortho/para-H2 ratio is in equilibrium (0) or normal 3:1 (1).
            iray: Sets the Rayleigh optical depth calculation:
                    0 = Rayleigh scattering optical depth is not included.
                    1 = Rayleigh optical depths suitable for gas giant atmosphere
                    2 = Rayleigh optical depths suitable for CO2 dominated atmosphere
                    3 = Rayleigh optical depths suitable for a N2-O2 atmosphere.
                    4 = New Rayleigh optical depths suitable for a gas giant atmosphere.
                    5 = New Rayleigh optical depths suitable for a gas giant atmosphere PLUS Raman scattering.
                    6 = New Rayleigh optical depths suitable for a gas giant atmosphere PLUS Sromovsky Raman scattering, PLUS Sromovsky Polarisation Correction.
            ih2o: Turns additional H2O continuum off (0) or on (1)
            ich4: Turns additional CH4 continuum off (0) or on (1)
            io3:  Turns additional O3 UV continuum off (0) or on (1)
            inh3: Turns additional NH3 continuum off (0) or on (1)
            iptf: Used in only a few routines to switch between normal partition function calculation (0) or the high-temperature partition function for CH4 for Hot Jupiters.
            imie: Only relevant for scattering calculations. If set to 0, the phase function is computed from the associated Henyey-Greenstein hgphase*.dat files. However, if set to 1, the phase function is computed from the MieTheory calculated PHASEN.DAT file.
        
        Returns:
            None
            
        Creates:
            nemesis.fla"""
        
        with open(self.directory+"nemesis.fla", mode="w+") as file:
            file.write('       {aa}	! Inormal (0=eqm, 1=normal)\n       {bb}	! Iray (0=off, 1=on)\n       {cc}	! IH2O (1 = turn extra continuum on)\n       {dd}	! ICH4 (1 = turn extra continuum on)\n       {ee}	! IO3 (1 = turn extra continuum on)\n       {ff}	! INH3 (1 = turn extra continuum on)\n       {gg}	! Iptf (0=default, 1=CH4 High-T)\n       {hh}	! IMie\n       {ii}	! UV Cross-sections\n       0	! INLTE (0=LTE)'.format(aa=inormal, bb=iray, cc=ih2o, dd=ich4, ee=io3, ff=inh3, gg=iptf, hh=imie, ii=iuv))

    def _generate_aerosol_ref(self):
        """Generate a number of aerosol reference profile that is 0 at all altitudes.
        
        Args:
            None

        Returns:
            None
            
        Creates:
            aerosol.ref"""
        
        # If there are no clouds defined return immediately
        if self.num_aerosol_modes == 0:
            return
        
        heights = self.ref.height
        with open(self.directory+"aerosol.ref", mode="w+") as file:
            file.write(f"# Generated by Eleos\n")
            file.write(f"{len(heights)} {self.num_aerosol_modes}\n")
            for h in heights:
                file.write(f"{h:>12.5f} ")
                for i in range(self.num_aerosol_modes):
                    file.write("0.000000E+00 ")
                file.write("\n")

    def _generate_kls(self):
        """Copy the ktables from the template core for the given instrument.
        Args:
            instrument: Either 'NIRSPEC' or 'MIRI'
            
        Returns:
            None
            
        Creates:
            nemesis.kls
            
        TODO:
            Add way to include/exclude different elements
            Add option to use ktables on ALICE rather than prepackaged"""
        if self.instrument == "NIRSPEC":
            shutil.copy(constants.PATH+"data/jupiter/nirspec.kls", self.directory+"nemesis.kls")
        elif self.instrument == "MIRI":
            shutil.copy(constants.PATH+"data/jupiter/miri.kls", self.directory+"nemesis.kls")

    def _generate_fmerror(self):
        """For each wavelength in the spx file, adjust the error by a factor (either fmerror_factor, _pct or _value) and write to the fmerror file. 
        Currently only supports 1 spx geometry
        Args:
            None
            
        Returns:
            None
        
        Creates:
            fmerror.dat """
        spx_data = spx.read(self.spx_file).geometries[0]
        num_entries = len(spx_data.wavelengths)

        if num_entries > 2048:
            raise IndexError(f"spx file has too many wavelengths! ({num_entries}/2048)")
        
        with open(self.directory + "fmerror.dat", mode="w+") as file:
            # Header with number of lines 
            file.write(f"{num_entries+2}\n")

            # Catch any cases outside the wavelength range (lower)
            file.write(f"{0.1:.6e}  {1e-8:.6e}\n")

            # Scale the spx error by either fmerror_factor, fmerror_pct or fmerror_value
            for wl, err, spc in zip(spx_data.wavelengths, spx_data.error, spx_data.spectrum):
                if self.fmerror_factor is not None:
                    file.write(f"{wl:.6e}  {err*self.fmerror_factor:.6e}\n")
                elif self.fmerror_pct is not None:
                    file.write(f"{wl:.6e}  {spc*self.fmerror_pct:.6e}\n")
                elif self.fmerror_value is not None:
                    file.write(f"{wl:.6e}  {self.fmerror_value:.6e}\n")

            # Catch any cases outside the wavelength range (upper)
            file.write(f"{100:.6e}  {1e-8:.6e}\n")

    def _generate_xsc(self):
        """Run Makephase and Normxsc with the given inputs to generate the .xsc and phase files.
        
        Args:
            None
            
        Returns:
            None
        
        Creates:
            makephase.inp
            normxsc.inp
            nemesis.xsc
            PHASE{N}.DAT
            hgphase{n}.dat"""
        
        # Check to see if any aerosols have been added - if not then return immediately
        if self.num_aerosol_modes == 0:
            return
        
        # Read in wavelengths bounds from spx file
        wls = spx.read(self.spx_file).geometries[0].wavelengths
        start_wl = min(wls) - 0.1
        end_wl = max(wls) + 0.1

        # Replace the number of aerosol modes and start/end/delta wavelengths
        with open(self.directory+"makephase.inp", mode="r+") as file:
            lines = file.read().split("\n")
            lines[0] = str(self.num_aerosol_modes)
            lines[2] = f"{start_wl} {end_wl} 0.1"
            file.seek(0)
            file.write("\n".join(lines))
            file.truncate()

        # Generate the normxsc.inp file
        with open(self.directory+"normxsc.inp", mode="w+") as file:
            file.write(f"nemesis.xsc\n1 1")

        # Run Makephase and Normxsc
        cwd = os.getcwd()
        os.chdir(self.directory)
        os.system("Makephase < makephase.inp")
        os.system("Normxsc < normxsc.inp")
        os.chdir(cwd)

    def _generate_fcloud_ref(self):
        """Generate the fcloud.ref file
        
        Args:
            None
            
        Returns:
            None
            
        Creates:
            fcloud.ref"""
        
        # Return immediately if there are no cloud layers defined
        if self.num_aerosol_modes == 0:
            return 
        
        with open(self.directory+"fcloud.ref", mode="w+") as file:
            file.write(f"{len(self.ref)}    {self.num_aerosol_modes}\n")
            for height in self.ref.height:
                file.write(f"      {height:> 9.4f}      {int(self.cloud_cover)}")
                for n in range(self.num_aerosol_modes):
                    file.write("      1")
                file.write("\n")

    def add_aerosol_mode(self, aerosol_profile=None, refractive_index_profile=None, radius=None, variance=None, refractive_index=None):
        """Add an aerosol mode. This uses the Mie scattering option with constant refractive index over the wavelength range, which is set to be the same as the input .spx files.
        Can be called with either of the following signatures:

        core.add_aerosol_mode(AerosolProfile(...), radius, variance, refractive_index)
        or
        core.add_aerosol_mode(AerosolProfile(...), ImagRefractiveIndexProfile(...))

        to pull parameters from an ImagRefractiveIndexProfile or specified manually respectively
        
        Args:
            refractive_index_profile (ImagRefractiveIndexProfile): If given, use the aerosol radius distribution and refractive index from the Profile specified
            radius (float): The mean radius of the aerosol particles
            variance (float): The variance of the particle size distribution
            refractive_index (complex): The aerosol's refractive index as a complex number (eg. 1.2 + 1e-3j)
            
        Returns:
            None"""

        # Increment aerosol mode counter        
        self.num_aerosol_modes += 1

        # Add the profile to the core
        self._add_profile(aerosol_profile)

        # Assign aerosol IDs
        aerosol_profile.aerosol_id = self.num_aerosol_modes

        # Pull parameters from the refractive index profile and add to the core if given
        if refractive_index_profile is not None:
            refractive_index_profile.aerosol_id = self.num_aerosol_modes
            refractive_index_profile.shape.aerosol_id = self.num_aerosol_modes
            self._add_profile(refractive_index_profile)

            radius = refractive_index_profile.shape.radius
            variance = refractive_index_profile.shape.variance
            refractive_index = refractive_index_profile.shape.refractive_index

        # Add the mode to the bottom of the Makephase input file
        with open(self.directory+"makephase.inp", mode="a") as file:
            file.write(f"1\n{radius} {variance}\n2\n1\n{refractive_index.real} {refractive_index.imag}\n")

    def generate_core(self):
        self._generate_inp()
        self._generate_set()
        self._generate_flags()
        self._generate_aerosol_ref()
        self._generate_kls()
        self._generate_fmerror()
        self._generate_fcloud_ref()
        self._generate_xsc()
        self._generate_apr()
        self._save_core()        


def parse_ref_file(ref_file):
    """Read in the .ref file provided and return a DataFrame
    
    Args:
        ref_file: Path to the .ref file
        
    Returns:
        pandas.DataFrame with columns for height (in km), pressure (in atm), temperature (in K), and the VMRs for all specified gases"""
    with open(ref_file) as file:
        for i, line in enumerate(file):
            if "height" in line:
                skip_to = i
            if i == 2:
                tokens = line.split()
                n_gas = int(tokens[-1])
    df = pd.read_table(ref_file, skiprows=skip_to+1, sep="\s+", names=["height", "pressure", "temp"] + [f"VMR gas {n+1}" for n in range(n_gas)])
    return df


def load_core(core_directory):
    """Load a NemesisCore object saved using NemesisCore._save_core
    
    Args:
        None
        
    Returns:
        NemesisCore: The unpickled core"""
    with open(core_directory+"core.pkl", 'rb') as f:
        return pickle.load(f)


def reset_core_numbering():
    """Reset the automatic core numbering. Useful for creating multiple sets of cores in one program.
    
    Args:
        None
        
    Returns:
        None"""
    NemesisCore.core_id = 0


def clear_parent_directory(parent_directory):
    """Attempt to delete all the child files/direcotries from the specified folder. It is recommended to call this function at the start
    of every core generation script as it ensures that the script is fully stateless. If a script generates N cores when first run
    and is subsequently modified to produce N/2 cores, then the remaining N/2 core directories will remain in the parent directory.
    This should not matter as the SLURM script will only run the correct subset of cores, but it can get confusing. If the parent directory
    specified does not exist then it will print a warning and return.
    
    Args:
        parent_directory (str): The directory to clear
        
    Returns:
        None"""
    try:
        shutil.rmtree(parent_directory)
    except FileNotFoundError:
        warnings.warn(f"Attempted to clear directory {parent_directory} but it does not exist.")
        return


def generate_alice_job(cores, username, memory=16, hours=24):
    """Generate an sbatch submission script for use on ALICE. The job is an array job over all the specified cores.
    
    Args:
        cores: Either a single core or a list of cores
        username: The username of the user running the job (eg, scat2, lnf2)
        memory: The amount of memory to use (in GB)
        hours: The number of hours to schedule the job for
        
    Returns:
        None
        
    Creates:
        submitjob.sh in the parent directory of the cores"""
    
    # If only one core is passed, make it into a 1-length list for consistency
    if isinstance(cores, NemesisCore):
        cores = [cores,]

    script_path = cores[0].parent_directory + "submitjob.sh"

    # Read the submission script and replace template fields
    with open(constants.PATH+"data/statics/template.job", mode="r") as file:
        out = file.read()
        out = out.replace("<MEMORY>", str(memory))
        out = out.replace("<HOURS>", f"{hours:02}")
        out = out.replace("<N_CORES>", str(len(cores)))
        out = out.replace("<CORE_DIR>", os.path.abspath(cores[0].parent_directory))
        out = out.replace("<USERNAME>", username)

    # Write the filled template 
    with open(script_path, mode="w+") as file:
        file.write(out)


def run_alice_job(parent_directory, print_queue_delay=2):
    """Run the script created by generate_alice_job
    
    Args:
        parent_directory (str): Path to the directory containing all the cores
        print_queue_delay (float): Duration to wait (in seconds) before printing the current job queue

    Returns:
        None
    """
    os.system(f"sbatch {parent_directory}submitjob.sh")
    time.sleep(print_queue_delay)
    os.system("squeue --me")